import os
from typing import List

from requests_futures.sessions import FuturesSession

from savvihub.api.file_object import UploadableFileObject, DownloadableFileObject
from savvihub.api.savvihub import SavviHubClient
from savvihub.api.types import SavviHubFileObject
from savvihub.common.utils import wait_all_futures, calculate_crc32c


def default_hooks():
    def fn(resp, **kwargs):
        resp.raise_for_status()
    return {
        'response': fn,
    }


class Downloader:
    @classmethod
    def get_files_to_download(cls, context, volume_id, *, snapshot='latest', prefix=''):
        client = SavviHubClient(token=context.token)
        files = client.volume_file_list(volume_id, prefix=prefix, snapshot=snapshot, recursive='true', raise_error='true')
        return files

    @classmethod
    def parallel_download(cls, context, base_path, files: List[SavviHubFileObject], progressable=None):
        if len(files) <= 0:
            return
        session = FuturesSession(max_workers=context.parallel)
        futures = []
        for file in files:
            if file.path.endswith('/'):
                continue

            d = DownloadableFileObject(file.download_url.get('url'), base_path, file.path)
            futures.append(d.download(session, progressable=progressable))

        wait_all_futures(futures)


class Uploader:
    @classmethod
    def get_files_to_upload(cls, base_path, hashmap=None):
        results = []
        for root, dirs, files in os.walk(base_path):
            for name in files:
                name = os.path.join(os.path.abspath(root), name)
                name = name[len(base_path) + 1:] if name.startswith(base_path) else name
                if hashmap and hashmap[name] == calculate_crc32c(os.path.join(base_path, name)):
                    continue
                results.append(name)
        return results

    @classmethod
    def get_hashmap(cls, base_path):
        files = cls.get_files_to_upload(base_path)
        hashmap = dict()
        for file in files:
            path = os.path.join(base_path, file)
            hashmap[file] = calculate_crc32c(path)
        return hashmap

    @classmethod
    def parallel_upload(cls, context, base_path, files, volume_id, *, progressable=None):
        if len(files) <= 0:
            return

        session = FuturesSession(max_workers=context.parallel)
        client = SavviHubClient(token=context.token, session=session)

        futures = []
        for file in files:
            future = client.volume_file_create(
                volume_id,
                file,
                is_dir=False,
                hooks=default_hooks()
            )
            futures.append(future)
        resps = wait_all_futures(futures)

        futures = []
        for i, resp in enumerate(resps):
            file_object = SavviHubFileObject(resp.json())

            u = UploadableFileObject(file_object.upload_url.get("url"), base_path, file_object.path)
            futures.append(u.upload(session, progressable=progressable))

        wait_all_futures(futures)

        futures = []
        for file in files:
            future = client.volume_file_uploaded(
                volume_id,
                file,
                hooks=default_hooks(),
            )
            futures.append(future)

        resps = wait_all_futures(futures)
        return [SavviHubFileObject(resp.json()) for resp in resps]



import requests
import typing

import typer

from savvihub.api.types import PaginatedMixin, SavviHubListResponse, SavviHubFileObject, SavviHubDataset, \
    SavviHubKernelResource, SavviHubKernelImage, SavviHubExperiment, SavviHubProject, SavviHubWorkspace, SavviHubUser, \
    SavviHubExperimentLog, SavviHubSnapshot, SavviHubVolume, SavviHubExperimentMetric
from savvihub.common.constants import API_HOST, WEB_HOST, DEBUG


class SavviHubClient:
    def __init__(self, *, session=requests.Session(), token=None, url=API_HOST, content_type='application/json'):
        self.session = session
        self.url = url
        self.token = token
        
        session.headers = {'content-type': content_type}
        if token:
            session.headers['authorization'] = 'Token %s' % token

    def get(self, url, params=None, raise_error=False, **kwargs):
        r = self.session.get(f'{self.url}{url}', params=params, **kwargs)
        if DEBUG:
            print({
                'method': 'GET',
                'url': url,
                'params': params,
                'status_code': r.status_code if isinstance(r, requests.Response) else None,
                'response': r.text if isinstance(r, requests.Response) else None,
            })
        if raise_error:
            r.raise_for_status()
        return r

    def get_all(self, url, params=None, raise_error=False, **kwargs):
        raw_r = self.get(url, params=params, raise_error=raise_error, **kwargs)
        r = PaginatedMixin(raw_r.json())
        results = []

        fetched_items = 0
        while True:
            fetched_items += len(r.results)
            results.extend(r.results)
            if fetched_items >= r.total:
                break
            raw_r = self.get(url, params={**params, 'after': r.endCursor}, raise_error=raise_error, **kwargs)
            r = PaginatedMixin(raw_r.json())
        return results

    def get_all_without_pagination(self, url, params=None, raise_error=False, **kwargs):
        raw_r = self.get(url, params=params, raise_error=raise_error, **kwargs)
        r = SavviHubListResponse(raw_r.json())
        return r.results

    def post(self, url, data, raise_error=False, **kwargs):
        r = self.session.post(f'{self.url}{url}', json=data, **kwargs)
        if DEBUG:
            print({
                'method': 'POST',
                'url': url,
                'request': data,
                'status_code': r.status_code if isinstance(r, requests.Response) else None,
                'response': r.text if isinstance(r, requests.Response) else None,
            })
        if raise_error:
            r.raise_for_status()
        return r

    def delete(self, url, raise_error=False, **kwargs):
        r = self.session.delete(f'{self.url}{url}', **kwargs)
        if DEBUG:
            print({
                'method': 'DELETE',
                'url': url,
                'status_code': r.status_code if isinstance(r, requests.Response) else None,
                'response': r.text if isinstance(r, requests.Response) else None,
            })
        if raise_error:
            r.raise_for_status()
        return r

    def patch(self, url, data, raise_error=False, **kwargs):
        r = self.session.patch(f'{self.url}{url}', json=data, **kwargs)
        if DEBUG:
            print({
                'method': 'PATCH',
                'url': url,
                'request': data,
                'status_code': r.status_code if isinstance(r, requests.Response) else None,
                'response': r.text if isinstance(r, requests.Response) else None,
            })
        if raise_error:
            r.raise_for_status()
        return r

    def get_my_info(self):
        r = self.get(f'/v1/api/accounts/me/')
        if r.status_code != 200:
            return None
        return SavviHubUser(r.json())

    def check_signin_token(self):
        r = self.post(f'/v1/api/accounts/signin/cli/token/', {})
        r.raise_for_status()
        return r.json()['cli_token']

    def check_signin(self, cli_token):
        r = self.get(f'/v1/api/accounts/signin/cli/check/', {
            'cli_token': cli_token,
        })
        r.raise_for_status()
        return r.json()['signin_success'], r.json()['access_token']

    def volume_create(self, workspace, role_type='experiment-input', **kwargs):
        r = self.post(f'/v1/api/volumes/', {
            'workspace_slug': workspace,
            'role_type': role_type,
        }, **kwargs)
        return SavviHubVolume(r.json())

    def volume_read(self, volume_id, **kwargs):
        r = self.get(f'/v1/api/volumes/{volume_id}/', **kwargs)
        return SavviHubVolume(r.json())

    def volume_file_list(self, volume_id, *, snapshot='latest', prefix='', recursive='false', **kwargs) \
            -> typing.List[SavviHubFileObject]:
        r = self.get(f'/v1/api/volumes/{volume_id}/files/',
                     params={'prefix': prefix, 'recursive': recursive, 'snapshot': snapshot}, **kwargs)
        return [SavviHubFileObject(x) for x in r.json().get('results')]

    def volume_file_read(self, volume_id, path, snapshot, **kwargs):
        r = self.get(f'/v1/api/volumes/{volume_id}/files/read/', params={'path': path, 'snapshot': snapshot}, **kwargs)
        if r.status_code != 200:
            return None
        return SavviHubFileObject(r.json())

    def volume_file_delete(self, volume_id, path, is_dir):
        r = self.delete(f'/v1/api/volumes/{volume_id}/files/delete/', params={'path': path, 'is_dir': is_dir})
        if r.status_code != 200:
            return None
        return [SavviHubFileObject(x) for x in r.json().get('results')]

    def volume_file_create(self, volume_id, path, is_dir, **kwargs):
        return self.post(f'/v1/api/volumes/{volume_id}/files/', {
            'path': path,
            'is_dir': is_dir
        }, **kwargs)

    def volume_file_uploaded(self, volume_id, path, **kwargs):
        return self.post(f'/v1/api/volumes/{volume_id}/files/uploaded/', {
            'path': path,
        }, **kwargs)

    def snapshot_read(self, volume_id, ref, **kwargs):
        r = self.get(f'/v1/api/volumes/{volume_id}/snapshots/{ref}/', **kwargs)
        if r.status_code != 200:
            return None
        return SavviHubSnapshot(r.json())

    def experiment_token_read(self, **kwargs):
        r = self.get(f'/v1/api/experiments/', **kwargs)
        if r.status_code != 200:
            return None
        return SavviHubExperiment(r.json())

    def experiment_read(self, workspace, project, experiment_number, **kwargs) -> SavviHubExperiment:
        return SavviHubExperiment(
            self.get(f'/v1/api/workspaces/{workspace}/projects/{project}/experiments/{experiment_number}',
                     **kwargs).json())

    def experiment_list(self, workspace, project, **kwargs):
        return [SavviHubExperiment(x)
                for x in self.get_all_without_pagination(
                f'/v1/api/workspaces/{workspace}/projects/{project}/experiments/?orderby.direction=DESC', **kwargs)]

    def experiment_log(self, workspace, project, experiment_number, **kwargs):
        return [SavviHubExperimentLog(x)
                for x in self.get(
                f'/v1/api/workspaces/{workspace}/projects/{project}/experiments/{experiment_number}/log/',
                **kwargs).json().get('logs')]

    def experiment_create(self, workspace, project, image_url, resource_spec_id, git_ref, git_diff_file_path,
                          input_volume_id, start_command, dataset_mount_infos, env_vars, **kwargs):
        return self.post(f'/v1/api/workspaces/{workspace}/projects/{project}/experiments/', {
            'image_url': image_url,
            'resource_spec_id': resource_spec_id,
            'git_ref': git_ref,
            'dataset_mount_infos': dataset_mount_infos,
            'git_diff_file_path': git_diff_file_path,
            'input_volume_id': input_volume_id,
            'env_vars': env_vars,
            'start_command': start_command,
        }, **kwargs)

    def experiment_metrics_update(self, experiment_id, metrics: typing.Dict[str, typing.List[SavviHubExperimentMetric]], **kwargs):
        return self.post(f'/v1/api/experiments/{experiment_id}/metrics/', {
            'metrics': metrics,
        }, **kwargs)

    def experiment_system_metrics_update(self, experiment_id, metrics: typing.Dict[str, typing.List[SavviHubExperimentMetric]], **kwargs):
        return self.post(f'/v1/api/experiments/{experiment_id}/system_metrics/', {
            'system_metrics': metrics,
        }, **kwargs)

    def kernel_image_list(self, workspace):
        results = self.get_all_without_pagination(f'/v1/api/workspaces/{workspace}/kernels/images/')
        return [SavviHubKernelImage(x) for x in results]

    def kernel_resource_list(self, workspace):
        results = self.get_all_without_pagination(f'/v1/api/workspaces/{workspace}/kernels/resource_specs')
        return [SavviHubKernelResource(x) for x in results]

    def workspace_read(self, workspace):
        r = self.get(f'/v1/api/workspaces/{workspace}/')
        if r.status_code == 404:
            return None
        elif r.status_code != 200:
            r.raise_for_status()
        return SavviHubWorkspace(r.json())

    def project_read(self, workspace, project):
        r = self.get(f'/v1/api/workspaces/{workspace}/projects/{project}/')
        if r.status_code == 404:
            return None
        elif r.status_code != 200:
            r.raise_for_status()
        return SavviHubProject(r.json())

    def project_github_create(self, workspace, project, github_owner, github_repo, **kwargs):
        return SavviHubProject(self.post(f'/v1/api/workspaces/{workspace}/projects_github/', {
            'slug': project,
            'github_owner': github_owner,
            'github_repo': github_repo,
        }, **kwargs).json())

    def public_dataset_list(self, **kwargs):
        results = self.get_all(f'/v1/api/datasets/public/', **kwargs)
        return [SavviHubDataset(x) for x in results]

    def dataset_list(self, workspace, **kwargs):
        results = self.get_all(f'/v1/api/workspaces/{workspace}/datasets/', **kwargs)
        if not results:
            return []
        return [SavviHubDataset(x) for x in results]

    def dataset_read(self, workspace, dataset, **kwargs):
        r = self.get(f'/v1/api/workspaces/{workspace}/datasets/{dataset}/', **kwargs)
        if r.status_code != 200:
            return None
        return SavviHubDataset(r.json())

    def dataset_snapshot_file_list(self, workspace, dataset, ref, **kwargs):
        r = self.get(f'/v1/api/workspaces/{workspace}/datasets/{dataset}/snapshots/{ref}/files/', **kwargs)
        return [SavviHubFileObject(x) for x in r.json().get('results')]

    def dataset_create(self, workspace, slug, is_public, description, **kwargs):
        r = self.post(f'/v1/api/workspaces/{workspace}/datasets/', {
            'slug': slug,
            'is_public': is_public,
            'description': description,
        }, **kwargs)
        if r.status_code == 409:
            typer.echo(f"Duplicate entity \"{workspace}/{slug}\". Try another dataset name!")
            return
        return SavviHubDataset(r.json())

    def dataset_gs_create(self, workspace, slug, is_public, description, gs_path, **kwargs):
        r = self.post(f'/v1/api/workspaces/{workspace}/datasets_gs/', {
            'slug': slug,
            'is_public': is_public,
            'description': description,
            'gs_path': gs_path
        }, **kwargs)
        if r.status_code == 409:
            typer.echo(f"Duplicate entity \"{workspace}/{slug}\". Try another dataset name!")
            return
        return SavviHubDataset(r.json())

    def dataset_s3_create(self, workspace, slug, is_public, description, s3_path, aws_role_arn, **kwargs):
        r = self.post(f'/v1/api/workspaces/{workspace}/datasets_s3/', {
            'slug': slug,
            'is_public': is_public,
            'description': description,
            's3_path': s3_path,
            'aws_role_arn': aws_role_arn,
        }, **kwargs)
        if r.status_code == 409:
            typer.echo(f"Duplicate entity \"{workspace}/{slug}\". Try another dataset name!")
            return
        return SavviHubDataset(r.json())

    @staticmethod
    def get_full_info(workspace, project, experiment, form):
        return \
            f'\n\tFull {form} info at:\n' \
            f'\t\t{WEB_HOST}/{workspace}/{project}/experiments/{experiment}/{form}\n'

    @staticmethod
    def get_full_info_project(workspace, project):
        return \
            f'\n\tFull project info at:\n' \
            f'\t\t{WEB_HOST}/{workspace}/{project}\n'

    @staticmethod
    def get_full_info_dataset(workspace, dataset):
        return \
            f'\n\tFull dataset info at:\n' \
            f'\t\t{WEB_HOST}/{workspace}/datasets/{dataset}\n'

    @staticmethod
    def get_full_info_experiment(workspace, project, experiment):
        return \
            f'\n\tFull experiment info at:\n' \
            f'\t\t{WEB_HOST}/{workspace}/{project}/experiments/{experiment}\n'

    @staticmethod
    def get_download_url(volume_id, path):
        return f'{WEB_HOST}/file-download/{volume_id}/{path}'



import inspect
import typeguard
import typing


class AnnotatedObject:
    def __init__(self, d: dict):
        resolved_types = {}
        for cls in self.__class__.mro():
            annotations = getattr(cls, '__annotations__', None)
            if annotations is None:
                continue

            for k, type_ in annotations.items():
                if k in resolved_types:
                    continue

                v = d.get(k, None)
                resolved_types[k] = type_

                if inspect.isclass(type_) and issubclass(type_, AnnotatedObject):
                    v = type_(v)
                else:
                    typeguard.check_type(k, v, type_)
                setattr(self, k, v)
        self.d = d

    def __str__(self):
        return f'[{self.__class__.__name__}] {self.d}'

    @property
    def dict(self):
        return self.d


class SavviHubUser(AnnotatedObject):
    id: int
    username: str
    name: str
    github_authorized: bool


class SavviHubWorkspace(AnnotatedObject):
    id: int
    slug: str


class SavviHubDatasetSource(AnnotatedObject):
    type: str
    support_snapshot: bool
    bucket_name: typing.Optional[str]
    prefix: typing.Optional[str]


class SavviHubDataset(AnnotatedObject):
    id: int
    workspace: SavviHubWorkspace
    slug: str
    main_volume_id: int
    source: SavviHubDatasetSource


class SavviHubFileObject(AnnotatedObject):
    path: str

    size: int
    hash: str

    download_url: typing.Optional[typing.Dict]
    upload_url: typing.Optional[typing.Dict]

    is_dir: typing.Optional[bool]


class SavviHubKernelImage(AnnotatedObject):
    id: int
    image_url: str
    name: str


class SavviHubKernelResource(AnnotatedObject):
    id: int
    name: str
    cpu_limit: float
    mem_limit: str


class SavviHubSnapshot(AnnotatedObject):
    id: int
    ref: str
    name: str
    size: int


class SavviHubProject(AnnotatedObject):
    id: int
    workspace: SavviHubWorkspace
    slug: str


class SavviHubExperiment(AnnotatedObject):
    class KernelImage(AnnotatedObject):
        name: str
        image_url: str
        language: str

    class KernelResourceSpec(AnnotatedObject):
        name: str
        cpu_type: str
        cpu_limit: int
        cpu_guarantee: int
        mem_limit: str
        mem_guarantee: str
        gpu_type: str
        gpu_limit: int
        gpu_guarantee: int

    id: int
    slug: str
    created_dt: str
    updated_dt: str
    number: int
    message: str
    status: str
    tensorboard_log_dir: typing.Optional[str]
    kernel_image: KernelImage
    kernel_resource_spec: KernelResourceSpec
    env_vars: typing.Optional[typing.Dict]
    datasets: typing.Optional[typing.List]
    tensorboard: typing.Optional[str]
    histories: typing.Optional[typing.List]
    output_volume_id: int
    metrics: typing.Optional[typing.List]
    source_code_link: str
    start_command: str


class SavviHubExperimentLog(AnnotatedObject):
    container: str
    stream: str
    message: str
    timestamp: float


class SavviHubExperimentMetric(AnnotatedObject):
    step: int
    value: str
    timestamp: float


class SavviHubListResponse(AnnotatedObject):
    results: typing.List


class PaginatedMixin(AnnotatedObject):
    total: int
    startCursor: typing.Optional[str]
    endCursor: typing.Optional[str]
    results: typing.List

class SavviHubVolume(AnnotatedObject):
    id: int
    created_dt: str
    updated_dt: str
    role_type: str
    workspace: typing.Optional[typing.Dict]
    project: typing.Optional[typing.Dict]
    dataset: typing.Optional[typing.Dict]
    experiment: typing.Optional[typing.Dict]



import concurrent.futures
import json
import os
import shutil

import zlib
from datetime import datetime

import timeago


def make_file(path):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    if not os.path.exists(path):
        with open(path, 'w'):
            print(" [*] Make a file : {}".format(path))
            pass


def make_dir(path):
    if not os.path.exists(path):
        print(" [*] Make directories : {}".format(path))
        os.makedirs(path)


def remove_file(path):
    if os.path.exists(path):
        print(" [*] Removed: {}".format(path))
        os.remove(path)


def remove_dir(path):
    if os.path.exists(path):
        print(" [*] Removed: {}".format(path))
        shutil.rmtree(path)


def write_json(path, data):
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(data, f, indent=4, sort_keys=True, ensure_ascii=False)


def load_json(path):
    with open(path) as f:
        data = json.loads(f.read())
    return data


def get_json_headers_with_token(token):
    try:
        authentication = 'token ' + token
    except TypeError:
        print('Token is not set!')
        return

    headers = {
        'Content-Type': 'application/json',
        'Authentication': authentication
    }

    return headers


def calculate_crc32c(filename):
    with open(filename, 'rb') as fh:
        h = 0
        while True:
            s = fh.read(65536)
            if not s:
                break
            h = zlib.crc32(s, h)
        return "%X" % (h & 0xFFFFFFFF)


def read_in_chunks(filename, blocksize=65535, chunks=-1, callback=None):
    """Lazy function (generator) to read a file piece by piece."""
    with open(filename, 'rb') as f:
        while chunks:
            data = f.read(blocksize)
            if not data:
                break
            yield data

            if callback:
                callback(data)
            chunks -= 1


def wait_all_futures(futures):
    concurrent.futures.wait(futures)
    return [future.result() for future in futures]


def is_committed():
    from git import Repo
    repo = Repo(search_parent_directories=True)
    changed_files = [item.a_path for item in repo.index.diff(None)]
    for file in changed_files:
        if file == 'savvihubfile.yml' or file == '.gitignore':
            continue
        else:
            return False
    return True

def short_string(data, length):
    return (data[:length] + '..') if len(data) > length else data


def parse_str_time_to_datetime(time_str):
    return datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S.%fZ")

def parse_time_to_ago(time_str):
    if not time_str:
        return 'None'
    time_datetime = parse_str_time_to_datetime(time_str)
    return timeago.format(time_datetime, datetime.now())

def parse_timestamp_or_none(timestamp):
    if not timestamp:
        return 'None'
    return datetime.fromtimestamp(timestamp)

def sizeof_fmt(num, suffix='B'):
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)



import os
from typing import Optional

from savvihub.api.types import SavviHubUser, SavviHubProject
from savvihub.common.git import GitRepository
from savvihub.common.config_loader import GlobalConfigLoader, ProjectConfigLoader
from savvihub.common.experiment import Experiment


class Value:
    def __init__(self, global_config_yml=None, savvihub_file_yml=None, env=None, computed=None, default_value=None):
        self.global_config_yml = global_config_yml
        self.savvihub_file_yml = savvihub_file_yml
        self.env = env
        self.computed = computed
        self.default_value = default_value


class Context:
    global_config = None
    project_config = None
    git_repo = None

    user: Optional[SavviHubUser] = None
    project: Optional[SavviHubProject] = None
    experiment: Optional[Experiment] = None

    experiment_token = os.environ.get("EXPERIMENT_ACCESS_TOKEN", None)
    parallel = os.environ.get("PARALLEL", 20)

    def __init__(self, login_required=False, project_required=False, experiment_required=False):
        if login_required:
            self.global_config = self.load_global_config()
            self.user = self.get_my_info()

            if self.user is None:
                if self.global_config.token:
                    raise Exception('Login required. You should call `savvi init` first.')
                else:
                    raise Exception('Token expired. You should call `savvi init` first.')

        if project_required:
            self.git_repo = GitRepository()
            self.project_config = self.load_project_config(self.git_repo.get_savvihub_config_file_path())
            self.project = self.get_project()

            if self.project is None:
                raise Exception('Project not found. Run `savvi project init`.')

        if experiment_required:
            self.experiment = self.get_experiment()

            if self.experiment is None:
                raise Exception('Experiment token required.')

    def get_my_info(self):
        from savvihub.api.savvihub import SavviHubClient
        client = SavviHubClient(token=self.global_config.token)
        return client.get_my_info()

    def get_experiment(self):
        from savvihub.api.savvihub import SavviHubClient
        client = SavviHubClient(token=self.experiment_token)

        experiment = client.experiment_token_read(raise_error=True)
        if experiment is None:
            return None

        return Experiment.from_given(experiment, client)

    def get_project(self):
        from savvihub.api.savvihub import SavviHubClient
        client = SavviHubClient(token=self.global_config.token)
        project = client.project_read(self.project_config.workspace, self.project_config.project)
        return project

    @property
    def authorized_client(self):
        from savvihub.api.savvihub import SavviHubClient
        return SavviHubClient(token=self.token)

    @property
    def token(self):
        if self.experiment_token:
            return self.experiment_token
        return self.global_config.token

    @staticmethod
    def load_global_config():
        return GlobalConfigLoader()

    @staticmethod
    def load_project_config(project_config_path):
        if not project_config_path:
            return None
        return ProjectConfigLoader(project_config_path)



from savvihub.common.context import Context


def log(step, row=None):
    """
    step: a step for each iteration (required)
    row: a dictionary to log
    """
    context = Context(experiment_required=True)
    context.experiment.log(step, row)
